{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "f7c52e6a",
   "metadata": {},
   "source": [
    "# Generative Adversarial Networks\n",
    "\n",
    "Thus far, we have discussed several **generative** models. A generative\n",
    "model learns the *structure* of a set of input data. In doing so, the model\n",
    "learns to *generate* new data that it has never seen before in the training\n",
    "data.\n",
    "\n",
    "A Generative Adversarial Network (GAN) is an example of a generative\n",
    "model."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "aba91eda",
   "metadata": {},
   "outputs": [],
   "source": [
    "%matplotlib inline\n",
    "\n",
    "import matplotlib.pyplot as plt\n",
    "import torch\n",
    "import torch.nn as nn\n",
    "import torch.nn.functional as F\n",
    "import torch.optim as optim\n",
    "import torch.utils.data\n",
    "from torchvision import datasets, transforms"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "b30e6622",
   "metadata": {},
   "source": [
    "We designed this tutorial to run on Google Colab. Go to \"Runtime\" and\n",
    "select \"Change Runtime Type\" and select \"GPU\". Run the next two lines\n",
    "of code to download the MNIST dataset."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "ac010157",
   "metadata": {},
   "outputs": [],
   "source": [
    "!wget www.di.ens.fr/~lelarge/MNIST.tar.gz\n",
    "!tar -zxvf MNIST.tar.gz"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "a0e67cf6",
   "metadata": {},
   "source": [
    "Then, let's begin by loading the MNIST training data."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "663dc84b",
   "metadata": {},
   "outputs": [],
   "source": [
    "mnist_data = datasets.MNIST('./', train=True, download=True, transform=transforms.ToTensor())"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "76326188",
   "metadata": {},
   "source": [
    "## Q1. Model\n",
    "\n",
    "A generative adversarial network (GAN) model consists of two models:\n",
    "\n",
    "- A **Generator** network $G$ that takes in a latent embedding (usually random noise)\n",
    "  and generates an image like those that exists in the training data\n",
    "- A **Discriminator** network $D$ that tries to distinguish between real images\n",
    "  from the training data, and fake images produced by the generator\n",
    "\n",
    "In essense, we have two neural networks that are adversaries: the generator\n",
    "wants to fool the discriminator, and the discriminator wants to avoid being\n",
    "fooled.\n",
    "\n",
    "Let's set up a simple generator and a discriminator to start:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "82eac0c9",
   "metadata": {},
   "outputs": [],
   "source": [
    "class Discriminator(nn.Module):\n",
    "    def __init__(self):\n",
    "        super().__init__()\n",
    "        self.model = nn.Sequential(\n",
    "            nn.Linear(28*28, 300),\n",
    "            nn.LeakyReLU(0.2, inplace=True),\n",
    "            nn.Linear(300, 100),\n",
    "            nn.LeakyReLU(0.2, inplace=True),\n",
    "            nn.Linear(100, 1)\n",
    "        )\n",
    "\n",
    "    def forward(self, x):\n",
    "        x = x.view(x.size(0), -1)\n",
    "        out = self.model(x)\n",
    "        return out.view(x.size(0))\n",
    "\n",
    "class Generator(nn.Module):\n",
    "    def __init__(self):\n",
    "        super().__init__()\n",
    "        self.model = nn.Sequential(\n",
    "            nn.Linear(100, 300),\n",
    "            nn.LeakyReLU(0.2, inplace=True),\n",
    "            nn.Linear(300, 28*28),\n",
    "            nn.Sigmoid()\n",
    "        )\n",
    "\n",
    "    def forward(self, x):\n",
    "        out = self.model(x).view(x.size(0), 1, 28, 28)\n",
    "        return out"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "7fe8dfe0",
   "metadata": {},
   "source": [
    "To make the GAN faster to train, both the Discriminator and Generator are\n",
    "fully-connected networks. You can use convolutional layers and other techniques\n",
    "we discussed.\n",
    "\n",
    "One difference between these models and the previous models we've built is\n",
    "that we are using a `nn.LeakyReLU` activation.\n",
    "Leaky ReLU is a variation\n",
    "of the ReLU activation that lets some information through, even when\n",
    "its input is less than 0. The layer `nn.LeakyReLU(0.2, inplace=True)` \n",
    "performs the computation: `x if x > 0 else 0.2 * x`.\n",
    "\n",
    "\n",
    "### Part (a)\n",
    "\n",
    "What tensor shape does the discriminator take as input?\n",
    "What does it produce as an output?"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "13a36e91",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Discuss the answers with your tutorial group"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "1d4b8205",
   "metadata": {},
   "source": [
    "### Part (b)\n",
    "\n",
    "What tensor shape does the generator take as input?\n",
    "What does it produce as an output?"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "3ce0ee89",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Discuss the answers with your tutorial group"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "e7c086fe",
   "metadata": {},
   "source": [
    "### Part (c)\n",
    "\n",
    "Explain what might be an advantage of using the leaky ReLU activation,\n",
    "compared to a ReLU activation."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "bc5b8998",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Discuss the answers with your tutorial group"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "ef94ef3e",
   "metadata": {},
   "source": [
    "## Q1. Training a GAN\n",
    "\n",
    "To train a GAN, we need to find a suitable loss function.\n",
    "Consider the following quantity:\n",
    "\n",
    "`P(D correctly identifies real image) + P(D correctly identifies image generated by G)`\n",
    "\n",
    "A good **discriminator** would want to *maximize* the above quanity by altering\n",
    "its parameters.\n",
    "\n",
    "Likewise, a good **generator** would want to *minimize* the above quanity. Actually,\n",
    "the only term that the generator controls is `P(D correctly identifies image generated by G)$`\n",
    "So, the best thing for the generator to do is alter its parameters to generate images\n",
    "that can fool D.\n",
    "\n",
    "Since we are looking at class probabilities, we will use binary cross entropy loss.\n",
    "\n",
    "Here is a rudimentary training loop to train a GAN.\n",
    "For every minimatch of data, we train the discriminator for one iteration,\n",
    "and then we train the generator for one iteration.\n",
    "\n",
    "For the discriminator, we use the label `1` to represent a **real** image, and `0` to represent\n",
    "a fake image.\n",
    "\n",
    "### Part (a)\n",
    "\n",
    "Fill in the training code below. The lines that you need to change is marked\n",
    "with a `#TODO` comment."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "5debd53b",
   "metadata": {},
   "outputs": [],
   "source": [
    "def train(generator, discriminator, device, lr=0.001, num_epochs=5):\n",
    "    # We will use the binary cross-entropy loss\n",
    "    criterion = nn.BCEWithLogitsLoss()\n",
    "\n",
    "    # From: https://colab.research.google.com/github/smartgeometry-ucl/dl4g/blob/master/gan.ipynb#scrollTo=-o7DOryPkUOb\n",
    "    # GAN training can be unstable. In this case, the strong momentum\n",
    "    # for the gradient prevents convergence. One possible explanation is that the\n",
    "    # strong momentum does not allow the two players in the adversarial game to react\n",
    "    # to each other quickly enough. Decreasing beta1 (the exponential decay for the\n",
    "    # gradient moving average in [0,1], lower is faster decay) from the default 0.9\n",
    "    # to 0.5 allows for quicker reactions.\n",
    "    d_optimizer = torch.optim.Adam(discriminator.parameters(), lr=lr, betas=(0.5, 0.999))\n",
    "    g_optimizer = torch.optim.Adam(generator.parameters(), lr=lr, betas=(0.5, 0.999))\n",
    "\n",
    "    # Use the MNIST data as training data\n",
    "    train_loader = torch.utils.data.DataLoader(mnist_data, batch_size=100, shuffle=True)\n",
    "\n",
    "    # We will track 16 noise vectors, and what image the generator creates\n",
    "    # from these 16 noise vectors.\n",
    "    num_test_samples = 16\n",
    "    test_noise = torch.randn(num_test_samples, 100).to(device)\n",
    "\n",
    "    # Move generator and discriminator weights to the GPU\n",
    "    generator = generator.to(device)\n",
    "    discriminator = discriminator.to(device)\n",
    "\n",
    "    for epoch in range(num_epochs):\n",
    "        # put the generator/discriminator in training mode\n",
    "        generator.train()\n",
    "        discriminator.train()\n",
    "\n",
    "        for n, (images, _) in enumerate(train_loader):\n",
    "            # === Train the Discriminator ===\n",
    "            noise = torch.randn(images.size(0), 100).to(device)\n",
    "            fake_images = generator(noise)\n",
    "            inputs = torch.cat([images.to(device), fake_images])\n",
    "            labels = None # TODO: Create a vector denoting that the real images\n",
    "                          #       should have label 1, and fake images should have label 0\n",
    "\n",
    "            d_outputs = discriminator(inputs)\n",
    "            d_loss = criterion(d_outputs, labels.to(device))\n",
    "            d_optimizer.zero_grad()\n",
    "            d_loss.backward()\n",
    "            d_optimizer.step()\n",
    "\n",
    "            # === Train the Generator ===\n",
    "            noise = torch.randn(images.size(0), 100)\n",
    "            fake_images = generator(noise.to(device))\n",
    "            outputs = discriminator(fake_images.to(device))\n",
    "\n",
    "            g_loss = criterion(outputs,\n",
    "                               None) # TODO: what should this be?\n",
    "            g_optimizer.zero_grad()\n",
    "            g_loss.backward()\n",
    "            g_optimizer.step()\n",
    "\n",
    "        # Report the average scores\n",
    "        scores = torch.sigmoid(d_outputs)\n",
    "        real_score = scores[:images.size(0)].data.mean()\n",
    "        fake_score = scores[images.size(0):].data.mean()\n",
    "\n",
    "        print('Epoch [%d/%d], d_loss: %.4f, g_loss: %.4f, ' \n",
    "              'D(x): %.2f, D(G(z)): %.2f' \n",
    "              % (epoch + 1, num_epochs, d_loss.item(), g_loss.item(), real_score, fake_score))\n",
    "        \n",
    "        # Plot images generated from the 16 noise vectors\n",
    "        generator.eval()\n",
    "        discriminator.eval()\n",
    "        test_images = generator(test_noise).cpu()\n",
    "        plt.figure(figsize=(9, 3))\n",
    "        for k in range(16):\n",
    "            plt.subplot(2, 8, k+1)\n",
    "            plt.imshow(test_images[k,:].data.numpy().reshape(28, 28), cmap='Greys')\n",
    "        plt.show()"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "9dee049e",
   "metadata": {},
   "source": [
    "### Part (b)\n",
    "\n",
    "Train the network for at least 20 epochs."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "d1b98b13",
   "metadata": {},
   "outputs": [],
   "source": [
    "discriminator = Discriminator()\n",
    "generator = Generator()\n",
    "\n",
    "device = torch.device(\"cuda:0\" if torch.cuda.is_available() else \"cpu\")\n",
    "#train(generator, discriminator, device, lr=0.0002, num_epochs=20)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "368dd76e",
   "metadata": {},
   "source": [
    "### Part (c)\n",
    "\n",
    "GANs are notoriously difficult to train. One difficulty is that a training curve is no longer\n",
    "as helpful as it was for a supervised learning problem! The generator and discriminator losses\n",
    "tend to bounce up and down, since both the generator and discriminator are changing over time.\n",
    "Tuning hyperparameters is also much more difficult, because we don't have the training curve\n",
    "to guide us. Newer GAN models like Wasserstein GAN tries to alleviate some of these issues, but\n",
    "are beyond the scope of this course.\n",
    "\n",
    "To compound the difficulty of hyperparameter tuning GANs also take a long time to train.\n",
    "It is tempting to stop training early, but the effects of hyperparameters may not be noticable\n",
    "until later on in training.\n",
    "\n",
    "Discuss what trend you notice in the discriminator and generator loss."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "78c22741",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Discuss the answers with your tutorial group"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "aa947edd",
   "metadata": {},
   "source": [
    "### Part (d)\n",
    "\n",
    "You might have noticed in the images generated by our simple GAN that the model seems to only\n",
    "output a small number of digit types.  This phenomenon is called **mode collapse**.\n",
    "Explain if mode collapse manifested in your GAN. If so, how? Why does mode collapse occur in\n",
    "a GAN?"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "00a201b6",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Discuss the answers with your tutorial group"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "2f2487a0",
   "metadata": {},
   "source": [
    "## Other GAN Implementations\n",
    "\n",
    "- https://github.com/prcastro/pytorch-gan/blob/master/MNIST%20GAN.ipynb\n",
    "- https://github.com/BeierZhu/GAN-MNIST-Pytorch\n",
    "- https://pytorch.org/tutorials/beginner/dcgan_faces_tutorial.html"
   ]
  }
 ],
 "metadata": {},
 "nbformat": 4,
 "nbformat_minor": 5
}
